/****************************************************************************
**
** This file is part of a Qt Solutions component.
** 
** Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).
** 
** Contact:  Qt Software Information (qt-info@nokia.com)
** 
** Commercial Usage  
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Solutions Commercial License Agreement provided
** with the Software or, alternatively, in accordance with the terms
** contained in a written agreement between you and Nokia.
** 
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
** 
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
** 
** GNU General Public License Usage 
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
** 
** Please note Third Party Software included with Qt Solutions may impose
** additional restrictions and it is the user's responsibility to ensure
** that they have met the licensing requirements of the GPL, LGPL, or Qt
** Solutions Commercial license and the relevant license of the Third
** Party Software they are using.
** 
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
** 
****************************************************************************/



/*!
    \class QtSequentialAnimationGroup
    \brief The QtSequentialAnimationGroup class provides a sequential group of animations.
    
    
    \preliminary

    The first animation in the group is started first, and when it stops, the next animation
    is started, and so on. The animation group finishes when the last animation has finished.

    You can call addPause() or insertPause() to add a pause to an
    animation.
*/

#include "qsequentialanimationgroup.h"
#include "qsequentialanimationgroup_p.h"

#include "qpauseanimation.h"

#include <QtCore/qdebug.h>

QT_BEGIN_NAMESPACE

QtSequentialAnimationGroupPrivate::IndexRangeForTime QtSequentialAnimationGroupPrivate::indexForTime(int msecs) const
{
    Q_Q(const QtSequentialAnimationGroup);
    Q_ASSERT(!animations.isEmpty());

    IndexRangeForTime ret;
    int duration = 0;

    // ### what if duration is -1
    ret.timeOffset = currentLoop * q->duration();

    for (int i = 0; i < animations.size(); ++i) {
        QtAbstractAnimation *animation = animations.at(i);
        duration = animation->totalDuration();
        if (duration == -1)
            duration = actualDuration.at(i); //we can try the actual duration there

        if (duration == 0) {
            if (ret.lower == -1)
                ret.lower = i;
            continue; //we'll find the next animation with a duration
        }

        //'animation' is the current animation if one of those reason is true
        // 1. it's duration is undefined
        // 2. it ends after msecs
        // 3. it is the last animations (this can happen in case there is at least 1 uncontrolled animation)
        if (duration == -1 || msecs <= (ret.timeOffset + duration)) {
            if (ret.lower == -1)
                ret.lower = i;
            ret.upper = i;
            //If we're at the end of the current animation we fast "forward" the animations which have a duration of 0
            if (msecs >= (ret.timeOffset + duration)) {
                while (ret.upper + 1 < animations.size()
                       && animations.at(ret.upper + 1)->totalDuration() == 0) {
                        ret.upper++;
                }
            }
            return ret;
        }

        //'animation' has a non-null defined duration and is not the one at time 'msecs'.
        ret.lower = -1; //put it back to uninitialized value
        ret.timeOffset += duration;
    }

    //this can only happen when one of those conditions is true
    //1. the duration of the group is undefined and we passed its actual duration
    //2. there are only 0-duration animations in the group
    ret.timeOffset -= duration;
    ret.upper = animations.size() - 1;
    if (ret.lower == -1)
        ret.lower = ret.upper;
    return ret;
}

/*!
    Constructs a QtSequentialAnimationGroup.
    \a parent is passed to QObject's constructor.
*/
QtSequentialAnimationGroup::QtSequentialAnimationGroup(QObject *parent)
    : QtAnimationGroup(*new QtSequentialAnimationGroupPrivate, parent)
{
}

/*!
    \internal
*/
QtSequentialAnimationGroup::QtSequentialAnimationGroup(QtSequentialAnimationGroupPrivate &dd,
                                                     QObject *parent)
    : QtAnimationGroup(dd, parent)
{
}

/*!
    Destroys the animation group. It will also destroy all its animations.
*/
QtSequentialAnimationGroup::~QtSequentialAnimationGroup()
{
}

/*!
    Adds a pause of \a msecs to this animation group.
    The pause is considered as a special type of animation, thus count() will be
    increased by one.
    \sa insertPauseAt(), add(), insertAt()
*/
QtPauseAnimation *QtSequentialAnimationGroup::addPause(int msecs)
{
    QtPauseAnimation *pause = new QtPauseAnimation(msecs);
    add(pause);
    return pause;
}

/*!
    Inserts \a animation into this animation group at \a index.
    If \a index is 0 the animation is inserted at the beginning.
    If \a index is count(), the animation is inserted at the end.
    \sa removeAt(), add(), animations()
*/
void QtSequentialAnimationGroup::insertAt(int index, QtAbstractAnimation *animation)
{
    Q_D(QtSequentialAnimationGroup);

    d->animations.insert(index, animation);
    if (QtAnimationGroup *oldGroup = animation->d_func()->group)
        oldGroup->remove(animation);
    animation->d_func()->group = this;
    animation->setParent(this);
    animationInsertedAt(index);
}

/*!
    Inserts a pause of \a msecs milliseconds at \a index in this animation
    group.

    \sa addPause(), insertAt(), animations()
*/
QtPauseAnimation *QtSequentialAnimationGroup::insertPauseAt(int index, int msecs)
{
    QtPauseAnimation *pause = new QtPauseAnimation(msecs);
    insertAt(index, pause);
    return pause;
}

/*!
    \reimp
*/
void QtSequentialAnimationGroup::animationInsertedAt(int index)
{
    Q_D(QtSequentialAnimationGroup);

    //### unfortunately we can't query the duration here because it is virtual pure in QtAbstractAnimation
    //-1 here means that the actual duration of the animation is not (yet) known
    //it can be changed when the animation stopped
    d->actualDuration.insert(index, -1);

    const int currentIndex = d->animations.indexOf(d->currentAnimation);

    if (d->currentAnimation == 0)
        d->setCurrentAnimation(0); // initialize the current animation

    if (index > currentIndex && d->currentLoop == 0)
        return; //we're not affected because it is added after the current one

    QtAbstractAnimation *newAnimation = d->animations.at(index);

    if ((currentIndex == index + 1) && (d->currentAnimation->currentTime() == 0)
        && (d->currentAnimation->currentLoop() == 0)) {
        //we're inserting an animation just where we are, so the current animation changes
        d->setCurrentAnimation(index);
        return;
    }

    const int lc = d->currentLoop + (currentIndex > index ? 1 : 0);
    if (lc) {
        const int animDuration = newAnimation->totalDuration();
        if (animDuration == -1)
            return; //we can't handle adding an animation with unknown duration

        //we take loopcount into account
        setCurrentTime(d->totalCurrentTime + lc * animDuration);
    }

}

/*!
    \reimp
*/
void QtSequentialAnimationGroup::animationRemovedAt(int index)
{
    Q_D(QtSequentialAnimationGroup);
    Q_ASSERT(d->currentAnimation); //currentAnimation should always be set

    //an animation was simply removed
    d->actualDuration.removeAt(index);

    const int currentIndex = d->animations.indexOf(d->currentAnimation);
    const int newIndex = (currentIndex == -1 ? index : currentIndex);
    if (currentIndex == -1) {
        //we're removing the current animation
        //let's update it to another animation
        if (index > 0) //try to take the previous one
            d->setCurrentAnimation(index-1);
        else if (index < d->animations.count())
            d->setCurrentAnimation(index);
        else
            d->currentAnimation = 0; // case all animations were removed
    }

    //duration of the previous animations up to the current animation
    int newCurrentTime = 0;
    for(int i = 0; i < newIndex; ++i) {
        const int current = d->animations.at(i)->totalDuration();
        if (current == -1)
            return; //we can't handle that
        newCurrentTime += current;
    }

    if (d->currentLoop) {
        //we make sure the current time is updated depending on loopcount
        const int dura = duration();
        if (dura == -1)
            return; //###we don't manage an group with -1 duration
        newCurrentTime += dura * d->currentLoop;
    }

    if (currentIndex != -1) {
        //the current animation is not the one being removed
        newCurrentTime += d->currentAnimation->d_func()->totalCurrentTime;
    }

    setCurrentTime(newCurrentTime);
}

/*!
    Returns the animation in the current time.

    \sa animations()
*/
QtAbstractAnimation *QtSequentialAnimationGroup::currentAnimation() const
{
    Q_D(const QtSequentialAnimationGroup);
    return d->currentAnimation;
}

/*!
    \reimp
*/
int QtSequentialAnimationGroup::duration() const
{
    Q_D(const QtSequentialAnimationGroup);
    int ret = 0;

    for (int i = 0; i < d->animations.size(); ++i) {
        QtAbstractAnimation *animation = d->animations.at(i);
        const int currentDuration = animation->totalDuration();
        if (currentDuration == -1)
            return -1; // Undetermined length

        ret += currentDuration;
    }

    return ret;
}

/*!
    \reimp
*/
void QtSequentialAnimationGroup::updateCurrentTime(int msecs)
{
    Q_D(QtSequentialAnimationGroup);
    if (!d->currentAnimation)
        return;

    const QtSequentialAnimationGroupPrivate::IndexRangeForTime newAnimationRange = d->indexForTime(msecs);
    const int currentAnimationIndex = d->animations.indexOf(d->currentAnimation);

    // newAnimationRange.upper is the new current animation. We need to make sure we execute
    // everything between newAnimationRange.lower and newAnimationRange.upper when starting
    // the new current animation.

    if (currentAnimationIndex != newAnimationRange.upper) {
        // ### when looping, ensure all following animations are completed,
        // then rewind (while stopped) and start over from the start.
        if (newAnimationRange.upper < currentAnimationIndex) {
            // let's rewind the animations (used when looping and seeking backwards)
            for (int i = currentAnimationIndex; i > newAnimationRange.upper; --i) {
                // ### this is also called if the anim is restarted, is that correct?
                // I suppose so (start after finish == restart / loop).
                if (i != currentAnimationIndex)
                    d->setCurrentAnimation(i);
                d->animations.at(i)->setCurrentTime(0);
                d->actualDuration[i] = -1;
            }
            // we need to reexecute the animations in order from lower to upper.
            // only lower can be a defined duration animation, all others will be 0 duration animations
            // ### what to do if lower is -1 duration?
            for (int i = newAnimationRange.lower; i < newAnimationRange.upper; ++i) {
                QtAbstractAnimation *anim = d->animations.at(i);
                d->setCurrentAnimation(i);
                anim->setCurrentTime(anim->totalDuration());
            }
        } else {
            // fast-forward the animations that were skipped since last call to updateCurrentTime
            for (int i = qMax(0, currentAnimationIndex); i < newAnimationRange.upper; ++i) {
                QtAbstractAnimation *anim = d->animations.at(i);
                // avoid starting the currentAnimation, which is already running
                if (i != currentAnimationIndex)
                    d->setCurrentAnimation(i);
                // we finish the animation
                int duration = anim->totalDuration();
                if (duration == -1)
                    duration = d->actualDuration.at(i);
                // ### duration can still be -1
                anim->setCurrentTime(duration);
            }
        }
        // let's set the actual new current animation here
        d->setCurrentAnimation(newAnimationRange.upper);
    }

    const int newCurrentTime = msecs - newAnimationRange.timeOffset;
    const int currentDuration = d->currentAnimation->totalDuration();

    //we try to detect if we're at the end of the group
    if (d->currentAnimation == d->animations.last()
        && currentDuration != -1 && newCurrentTime >= currentDuration) {
        const int delta = newCurrentTime - d->currentAnimation->totalDuration();
        //here we've detected that the currenttime is beyond the end of the group
        d->currentTime -= delta;
        d->currentAnimation->setCurrentTime(d->currentAnimation->totalDuration());
        //### we should actually 'just' finish the current loop and not stop immediately
        stop();
    } else {
        d->currentAnimation->setCurrentTime(msecs - newAnimationRange.timeOffset);
    }
}

/*!
    \reimp
*/
void QtSequentialAnimationGroup::updateState(QtAbstractAnimation::State oldState,
                                            QtAbstractAnimation::State newState)
{
    Q_D(QtSequentialAnimationGroup);
    QtAnimationGroup::updateState(oldState, newState);

    if (!d->currentAnimation)
        return;

    switch (newState) {
    case NotRunning:
        for (int i = 0; i < d->animations.size(); ++i)
            d->actualDuration[i] = -1;
        d->currentAnimation->stop();
        break;
    case Paused:
        if (oldState == d->currentAnimation->state()
            && oldState == QtSequentialAnimationGroup::Running)
            d->currentAnimation->pause();
        else
            d->setCurrentAnimation(d->animations.indexOf(d->currentAnimation));
        break;
    case Running:
        if (oldState == d->currentAnimation->state()
            && oldState == QtSequentialAnimationGroup::Paused)
            d->currentAnimation->start();
        else
            d->setCurrentAnimation(d->animations.indexOf(d->currentAnimation));
        break;
    }
}

/*!
    \reimp
*/
void QtSequentialAnimationGroup::updateDirection(QtAbstractAnimation::Direction direction)
{
    Q_D(QtSequentialAnimationGroup);
    // we need to update the direction of the current animation
    if (state() != NotRunning && d->currentAnimation)
        d->currentAnimation->setDirection(direction);
}

void QtSequentialAnimationGroupPrivate::setCurrentAnimation(int index)
{
    Q_Q(QtSequentialAnimationGroup);

    // stop the old current animation
    if (currentAnimation)
        currentAnimation->stop();
    currentAnimation = animations.at(index);

    if (state == QtSequentialAnimationGroup::NotRunning)
        return;

    currentAnimation->stop();

    // we ensure the direction is consistent with our direction
    currentAnimation->setDirection(direction);

    // connects to finish signal of uncontrolled animations
    const int totalDuration = currentAnimation->totalDuration();
    if (totalDuration == -1)
        QObject::connect(currentAnimation, SIGNAL(finished()), q, SLOT(_q_uncontrolledAnimationFinished()));

    if (state == QtSequentialAnimationGroup::Running)
        currentAnimation->start();
    else
        currentAnimation->pause();
}

void QtSequentialAnimationGroupPrivate::_q_uncontrolledAnimationFinished()
{
    Q_Q(QtSequentialAnimationGroup);
    Q_ASSERT(qobject_cast<QtAbstractAnimation *>(q->sender()) == currentAnimation);

    const int index = animations.indexOf(currentAnimation);

    // we trust the duration returned by the animation
    actualDuration[index] = currentAnimation->currentTime();

    QObject::disconnect(currentAnimation, SIGNAL(finished()), q, SLOT(_q_uncontrolledAnimationFinished()));

    if (currentAnimation == animations.last()) {
        // ### we don't handle looping here
        q->stop();
    } else {
        int index = animations.indexOf(currentAnimation);
        if (index != -1) {
            // ### update internal timer offset
            setCurrentAnimation(index + 1);
        }
    }
}

QT_END_NAMESPACE

#include "moc_qsequentialanimationgroup.cpp"

